"""sc-githooks - Checks on Git commit lists

Copyright (c) 2021 Scott Lau
Portions Copyright (c) 2021 InnoGames GmbH
Portions Copyright (c) 2021 Emre Hasegeli
"""

from time import time

from githooks.base_check import BaseCheck, Severity
from githooks.git import CommitList, Commit


class CommitListCheck(BaseCheck):
    """Parent class for all commit list checks"""
    commit_list = None

    def prepare(self, obj):
        if not isinstance(obj, CommitList):
            return super(CommitListCheck, self).prepare(obj)

        new = self.clone()
        new.commit_list = obj
        return new

    def __str__(self):
        return '{} 位于提交 {}'.format(type(self).__name__, self.commit_list)


class CheckDuplicateCommitSummaries(CommitListCheck):
    """Check repeated commit summaries on a single commit list

    We are not exact matching the commit summaries, but searching summaries
    on the beginning of other ones.  This covers summaries like "Fix the bug"
    and "Fix the bug really" which is common bad practice for some reason.
    """
    def prepare(self, obj):
        if isinstance(obj, CommitList) and len(obj) <= 1:
            return None
        return super(CheckDuplicateCommitSummaries, self).prepare(obj)

    def get_problems(self):
        duplicate_summaries = [()]  # Nothing starts with an empty tuple.
        for commit in sorted(self.commit_list, key=Commit.get_summary):
            summary = commit.get_summary()
            if summary.startswith(duplicate_summaries[0]):
                duplicate_summaries.append(summary)
                continue
            if len(duplicate_summaries) > 1:
                yield (
                    Severity.ERROR, '提交标题 "{}" 重复了 {} 次'.format(
                        min(duplicate_summaries, key=len),
                        len(duplicate_summaries),
                    )
                )
            duplicate_summaries = [summary]


class CheckMisleadingMergeCommit(CommitListCheck):
    merge_template = "Merge branch '{}'"

    def get_problems(self):
        branch_name = self.commit_list.branch_name
        for commit in self.commit_list:
            summary = commit.get_summary()
            if summary.startswith(self.merge_template.format(branch_name)):
                yield Severity.WARNING, '合并提交到自身分支'
            elif summary.startswith(self.merge_template.format('master')):
                yield Severity.WARNING, '合并提交到master分支'


class CheckTimestamps(CommitListCheck):
    current_timestamp = time()

    def get_problems(self):
        previous_author_timestamp = 0
        previous_committer_timestamp = 0
        for commit in self.commit_list:
            author_timestamp = commit.get_author().timestamp
            committer_timestamp = commit.get_committer().timestamp
            if author_timestamp > self.current_timestamp + 2:
                yield (
                    Severity.ERROR,
                    '提交 {} 的作者的时间是未来时间'
                    .format(commit),
                )
            if committer_timestamp > self.current_timestamp + 2:
                yield (
                    Severity.ERROR,
                    '提交 {} 的提交者的时间是未来时间'
                    .format(commit),
                )
            if author_timestamp > committer_timestamp:
                yield (
                    Severity.ERROR,
                    '提交 {} 的作者的时间在提交者的时间之后'
                    .format(commit),
                )
            if previous_author_timestamp > author_timestamp:
                yield (
                    Severity.NOTICE,
                    '提交 {} 的作者时间在前一个提交的时间之前'
                    .format(commit),
                )
            if previous_committer_timestamp > committer_timestamp:
                yield (
                    Severity.ERROR,
                    '提交 {} 的提交者时间在前一个提交的时间之前'
                    .format(commit),
                )
            previous_author_timestamp = author_timestamp
            previous_committer_timestamp = committer_timestamp


class CheckContributors(CommitListCheck):
    """Validate consistency of committer and author name and email addresses

    We threat committers and authors the same way in this class.  It is common
    Git bad practice to commit with different combinations of names and
    email addresses ruining useful statistics that can be made using the Git
    history.  There is not too much we can do about without having
    the possible list of all users, and that is certainly not something we
    would like.

    In this class, we are searching and indexing the Git commits in the past
    to find out the same names and email address, and cross check them with
    the current commits.  We are using the name together with the domain part
    of the email address.  It is common for some systems to commit changes
    in behalf of the user with a different email address.  Including
    the domain on the index would let it happen.
    """
    def prepare(self, obj):
        new = super(CheckContributors, self).prepare(obj)
        if new and isinstance(obj, CommitList):
            new.email_index = {}
            new.domain_index = {}
            new.name_index = {}

        return new

    def get_problems(self):
        old_contributors = self.get_old_contributors()
        for commit in self.commit_list:
            for contributor in commit.get_contributors():
                found = self.index_contributors(old_contributors, contributor)

                for problem in self.check_contributor(contributor, commit):
                    yield problem

                    # If there are any problems, evidently the contributor
                    # is not consistent with the indexes.  We override
                    # the indexes after reporting problem to avoid the same
                    # ones to be reported again.
                    found = False

                if not found:
                    self.index_contributor(contributor, override=True)

    def get_old_commits(self):
        """Yield old commits in reverse order

        We could call "git rev-list" here, but it is easy enough to implement
        the same using the content we already need for other reasons.  Though
        "git rev-list" would order the commits nicer, we are not putting any
        effort to the ordering in here as our caller is not sensitive.
        """
        unused_commits = self.commit_list[0].get_parents()
        commit_ids = {c.commit_id for c in unused_commits}
        while unused_commits:
            unused_commit = unused_commits.pop(0)
            yield unused_commit
            for commit in unused_commit.get_parents():
                if commit.commit_id not in commit_ids:
                    unused_commits.append(commit)
                    commit_ids.add(commit.commit_id)

    def get_old_contributors(self):
        for commit in self.get_old_commits():
            for contributor in commit.get_contributors():
                yield contributor

    def index_contributor(self, contributor, override=False, dry_run=False):
        """Index a single contributor

        The function does nothing when the contributor is already indexed.
        It returns true when if would add the contributor to any index.
        The dry_run argument is used to test only, if the contributor is
        indexed.
        """
        found = False

        if contributor.email not in self.email_index or override:
            if not dry_run:
                self.email_index[contributor.email] = contributor
            found = True

        domain = contributor.get_email_domain()
        if domain not in self.domain_index or override:
            if not dry_run:
                self.domain_index[domain] = None
            found = True

        if (contributor.name, domain) not in self.name_index or override:
            if not dry_run:
                self.name_index[(contributor.name, domain)] = contributor
            found = True

        return found

    def index_contributors(self, contributors, searched):
        """Index contributors until the searched one is found

        The function stops when the searched one is found.  We assume
        the caller to pass the existing generator again to avoid starting over
        to search the next one.  It returns with true when the searched one
        is found; returns with false when all of the items are indexes but
        the searched one is not found.
        """
        while self.index_contributor(searched, dry_run=True):
            for contributor in contributors:
                if self.index_contributor(contributor):
                    break
            else:
                return False
        return True

    def check_contributor(self, contributor, commit):
        """Check one contributor against the indexes"""

        other = self.email_index.get(contributor.email)
        if other and contributor.name != other.name:
            yield (
                Severity.ERROR,
                '提交 {} 的作者使用的名字 "{}" 而不是 "{}", 而两者的邮箱是一致的'
                .format(commit, contributor.name, other.name),
            )

        domain = contributor.get_email_domain()
        if domain not in self.domain_index:
            yield (
                Severity.NOTICE,
                '提交 {} 的作者邮箱是一个新域名 "{}"'
                .format(commit, domain),
            )

        other = self.name_index.get((contributor.name, domain))
        if other and contributor.email != other.email:
            yield (
                Severity.ERROR,
                '提交 {} 的作者使用了一个不同的邮箱 "{}" 而之前同名作者的邮箱是 "{}"'
                .format(commit, contributor.email, other.email),
            )



# -*- coding: utf-8 -*- 
# @Time : 4/12/21 3:35 PM 
# @Author : mxt
# @File : default.py
DEFAULT_CONFIG = {
    # 是否开发模式
    "dev": {
        "dev_mode": False,
    },
    "commit_check": {
        # 是否启动提交检查
        "enabled": True,
        # 提交标题行最大长度
        "commit_summary_max_length": 50,
        # 提交详细行最大行长度
        "commit_line_max_length": 80,

        # 二进制文件类型黑名单（逗号分隔的后缀名清单，无需带.，通过后缀名控制）
        "binary_file_illegal_suffixes": "jar,doc,docx",
        # 二进制文件类型白名单（逗号分隔的后缀名清单，无需带.，通过后缀名控制）
        "binary_file_legitimate_suffixes": "a,framework,so",
        # 二进制文件白名单（逗号分隔的区分大小写的文件名列表，这些二进制文件允许提交）
        "legal_binary_filenames": "gradle-wrapper.jar,maven-wrapper.jar",

        # 提交文件最大大小(10MB)
        "commit_file_max_size": 10485760,
    }
}




"""sc-githooks - Pre-receive hook routines

Copyright (c) 2021 MaoXinTeng
Copyright (c) 2021 Scott Lau
Portions Copyright (c) 2021 InnoGames GmbH
Portions Copyright (c) 2021 Emre Hasegeli
"""

from fileinput import input
from sys import stdout, stderr
from traceback import print_exc

from githooks.base_check import CheckState, prepare_checks, BaseCheck
from githooks.checks import checks
from githooks.config import config
from githooks.git import Commit
from githooks.utils import iter_buffer


class Runner(object):
    def __init__(self):
        self.checked_commit_ids = set()
        self.changed_file_checks = dict()

    def run(self):
        # check whether commit check is enabled
        commit_check_is_enabled = False
        try:
            commit_check_is_enabled = config.get("commit_check.enabled")
        except:
            pass
        if not commit_check_is_enabled:
            return CheckState.DONE

        state = CheckState.NEW

        # 我们正在缓冲检查，让它们在后台并行运行。并行化只适用于CheckCommands。
        # 它没有开销，因为我们必须以同样的方式在外部运行这些命令。
        # 我们只有一个限制，以避免消耗太多的进程。
        # add DEV_MODE
        dev_mode = False
        try:
            dev_mode = config.get("dev.dev_mode")
        except:
            pass
        if dev_mode:
            for check in self.expand_checks(checks):
                check.print_problems()
                assert check.state >= CheckState.DONE
                state = max(state, check.state)
        else:
            for check in iter_buffer(self.expand_checks(checks), 16):
                check.print_problems()
                assert check.state >= CheckState.DONE
                state = max(state, check.state)
        return state

    def expand_checks(self, _checks):
        next_checks = []
        for check in prepare_checks(_checks, None, next_checks):
            yield check

        for line in input():
            for check in self.expand_checks_to_input(next_checks, line):
                yield check

    def expand_checks_to_input(self, _checks, line):
        line_split = line.split()
        ref_path_split = line_split[2].split('/', 2)
        if ref_path_split[0] != 'refs' or len(ref_path_split) != 3:
            # We have no idea what this is.
            return

        commit = Commit(line_split[1])
        if not commit:
            # This is a deletion.  We don't check anything on deletes.
            return

        if ref_path_split[1] == 'heads':
            name = line_split[2]
            for check in self.expand_checks_to_branch(_checks, commit, name):
                yield check
        elif ref_path_split[1] == 'tags':
            for check in self.expand_checks_to_commit(_checks, commit):
                yield check

    def expand_checks_to_branch(self, _checks, commit, name):
        commit_list = commit.get_new_commit_list(name)

        # Appending the actual commit on the list to the new ones makes
        # testing easier.
        if commit not in commit_list:
            commit_list.append(commit)

        for check in self.expand_checks_to_commit_list(_checks, commit_list):
            yield check

    def expand_checks_to_commit_list(self, _checks, commit_list):
        next_checks = []
        for check in prepare_checks(_checks, commit_list, next_checks):
            yield check

        for commit in commit_list:
            if commit.commit_id not in self.checked_commit_ids:
                for check in self.expand_checks_to_commit(next_checks, commit):
                    yield check
                self.checked_commit_ids.add(commit.commit_id)

    def expand_checks_to_commit(self, _checks, commit):
        next_checks = []
        for check in prepare_checks(_checks, commit, next_checks):
            yield check

        for changed_file in commit.get_changed_files():
            for check in self.expand_checks_to_file(next_checks, changed_file):
                yield check

    def expand_checks_to_file(self, _checks, changed_file):
        # We first need to wait for the previous checks on the same file
        # to finish.  If one of them failed already, we don't bother checking
        # the same file again.  The committer should return back to her
        # commit she broke the file.  It makes too much noise to complain
        # about the same file on multiple commits.
        previous_checks = self.changed_file_checks.setdefault(
            changed_file.path, []
        )
        for check in previous_checks:
            assert check.state >= CheckState.CLONED
            # Wait for the check to run
            while check.state < CheckState.DONE:
                yield None
            if check.state >= CheckState.FAILED:
                return

        for check in prepare_checks(_checks, changed_file):
            yield check
            previous_checks.append(check)


def main():
    try:
        state = Runner().run()
    except Exception as e:
        # Flush the problems we have printed so far to avoid the traceback
        # appearing in between them.
        stdout.flush()
        print(file=stderr)
        print('{} An error occurred: {}'.format(BaseCheck.ERROR_MSG_PREFIX % "ERROR", e), file=stderr)
        print_exc()
        return 1
    else:
        if state >= CheckState.FAILED:
            return 1
    return 0



"""Configurations

Copyright (c) 2021 Scott Lau
"""

from sys import stdout, stderr
from traceback import print_exc

from scconfig.config import Config

from githooks.base_check import BaseCheck
from githooks.default import DEFAULT_CONFIG

# =========================================
#       INSTANCES
# --------------------------------------
try:
    config = Config.create(project_name="sc-githooks", defaults=DEFAULT_CONFIG)
except Exception as e:
    stdout.flush()
    print(file=stderr)
    print('{} failed to read configuration: {}'.format(BaseCheck.ERROR_MSG_PREFIX % "ERROR", e), file=stderr)
    print_exc()
